'use strict';

Object.defineProperty(exports, '__esModule', { value: true });

var mitt = require('mitt');
var minDash = require('min-dash');
var preact = require('preact');
var hooks = require('preact/hooks');
var snarkdown = require('snarkdown');
var jsxRuntime = require('preact/jsx-runtime');
var Markup = require('preact-markup');
var compat = require('preact/compat');

function _interopDefaultLegacy (e) { return e && typeof e === 'object' && 'default' in e ? e : { 'default': e }; }

var mitt__default = /*#__PURE__*/_interopDefaultLegacy(mitt);
var snarkdown__default = /*#__PURE__*/_interopDefaultLegacy(snarkdown);
var Markup__default = /*#__PURE__*/_interopDefaultLegacy(Markup);

class Validator {
  validateField(field, value) {
    const {
      validate
    } = field;
    let errors = [];

    if (!validate) {
      return errors;
    }

    if (validate.pattern && value && !new RegExp(validate.pattern).test(value)) {
      errors = [...errors, `Field must match pattern ${validate.pattern}.`];
    }

    if (validate.required && typeof value === 'undefined') {
      errors = [...errors, 'Field is required.'];
    }

    if ('min' in validate && value && value < validate.min) {
      errors = [...errors, `Field must have minimum value of ${validate.min}.`];
    }

    if ('max' in validate && value && value > validate.max) {
      errors = [...errors, `Field must have maximum value of ${validate.max}.`];
    }

    if ('minLength' in validate && value && value.trim().length < validate.minLength) {
      errors = [...errors, `Field must have minimum length of ${validate.minLength}.`];
    }

    if ('maxLength' in validate && value && value.trim().length > validate.maxLength) {
      errors = [...errors, `Field must have maximum length of ${validate.maxLength}.`];
    }

    return errors;
  }

}

function findData(data, path) {
  for (const key of path) {
    data = data[key];

    if (data === undefined) {
      return undefined;
    }
  }

  return data;
}
function findErrors(errors, path) {
  return errors[pathStringify(path)];
}
function findFieldRenderer(renderers, type) {
  return renderers.find(renderer => {
    return renderer.type === type;
  });
}
function isRequired(field) {
  return field.required;
}
function pathParse(path) {
  if (!path) {
    return [];
  }

  return path.split('.').map(key => {
    return isNaN(parseInt(key)) ? key : parseInt(key);
  });
}
function pathsEqual(a, b) {
  return a && b && a.length === b.length && a.every((value, index) => value === b[index]);
}
function pathStringify(path) {
  if (!path) {
    return '';
  }

  return path.join('.');
}
const indices = {};

const generateIndexForType = type => {
  if (type in indices) {
    indices[type]++;
  } else {
    indices[type] = 1;
  }

  return indices[type];
};

const generateIdForType = type => {
  return `${type}${generateIndexForType(type)}`;
};
function clone(data, replacer) {
  return JSON.parse(JSON.stringify(data, replacer));
}
function importSchema(schema) {
  schema = clone(schema);
  const fields = new Map();
  importField(schema, fields);
  return {
    schema,
    fields
  };
}

function importField(field, fields, parent, index) {
  const id = generateIdForType(field.type);
  field.id = id;
  fields.set(id, field);

  if (parent) {
    field.parent = parent.id;
    field.path = parent.path ? [...parent.path, 'components', index] : ['components', index];
  } else {
    field.path = [];
  }

  if (field.components) {
    importFields(field.components, fields, field);
  }
}

function importFields(components, fields, parent) {
  components.forEach((component, index) => {
    importField(component, fields, parent, index);
  });
}

/**
 * @typedef { any } Schema
 * @typedef { { [x: string]: any } } Data
 * @typedef { { [x: string]: string[] } } Errors
 * @typedef { ('readOnly') } Properties
 * @typedef { { readOnly?: boolean } } PropertyOptions
 * @typedef { ('submit' | 'changed') } Events
 * @typedef { { data: Data, schema: Schema, properties?: PropertyOptions } } FormCoreOptions
 * @typedef { { data: Data, errors: Errors, schema: Schema, properties: PropertyOptions } } State
 */

/**
 * The form core.
 */

class FormCore {
  /**
   * @constructor
   * @param { FormCoreOptions } options
   */
  constructor(options) {
    const {
      schema = {},
      data = {},
      properties = {}
    } = options;
    /**
     * @private
     */

    this.emitter = mitt__default['default']();
    this.validator = new Validator();
    this.fields = new Map();
    const {
      schema: importedSchema
    } = importSchema(schema);
    /**
     * @private
     * @type {Schema}
     */

    this.initialSchema = clone(importedSchema);
    /**
     * @private
     * @type {Data}
     */

    this.initialData = clone(data);
    /**
     * @type {State}
     */

    this.state = {
      data: clone(this.initialData),
      errors: {},
      schema: clone(this.initialSchema),
      properties
    };
  }

  reset() {
    this.emitter.emit('reset');
    this.setState({
      data: clone(this.initialData),
      schema: clone(this.initialSchema),
      errors: {}
    });
  }
  /**
   * @returns { { data: Data, errors: Errors } }
   */


  submit() {
    const data = this.state.data;
    const errors = this.validateAll(data);
    const payload = clone({
      data,
      errors
    });
    this.emitter.emit('submit', payload);
    return payload;
  }
  /**
   * @param  { { path: (string|number)[], value: any } } update
   */


  update(update) {
    const {
      path,
      value
    } = update;
    const field = Array.from(this.fields.values()).find(field => pathsEqual(field.path, path));
    const fieldErrors = this.validator.validateField(field, value);
    const data = minDash.set(this.getState().data, path, value);
    const errors = minDash.set(this.getState().errors, path, fieldErrors.length ? fieldErrors : undefined);
    this.setState({
      data,
      errors
    });
  }
  /**
   * @param { Data } data
   * @returns Errors
   */


  validateAll(data) {
    const errors = Array.from(this.fields.values()).reduce((errors, field) => {
      const {
        path
      } = field;
      const value = findData(data, path);
      const fieldErrors = this.validator.validateField(field, value);
      return minDash.set(errors, path, fieldErrors.length ? fieldErrors : undefined);
    }, {});
    this.setState({
      errors
    });
    return this.getState().errors;
  }

  getState() {
    return clone(this.state);
  }
  /**
   * @param { Partial<State> } state
   */


  setState(state) {
    this.state = { ...this.state,
      ...state
    };
    this.changed(this.state);
  }
  /**
   * @param { State } state
   */


  changed(state) {
    this.emitter.emit('changed', clone(state));
  }
  /**
   * @param { Properties } property
   * @param { any } value
   */


  setProperty(property, value) {
    const properties = minDash.set(this.getState().properties, [property], value);
    this.setState({
      properties
    });
  }
  /**
   * @type { (event: Events, callback: (state: any) => void) => void }
   */


  on(event, callback) {
    this.emitter.on(event, callback);
  }
  /**
   * @type { (event: Events, callback: (state: any) => void) => void }
   */


  off(event, callback) {
    this.emitter.off(event, callback);
  }

}

const FormRenderContext = preact.createContext({
  Empty: props => {
    return null;
  },
  Children: props => {
    return props.children;
  },
  Element: props => {
    return props.children;
  }
});

const FormContext = preact.createContext({
  fields: new Map(),
  data: {},
  errors: {},
  properties: {},
  schema: {},

  getFieldRenderer(type) {
    return null;
  }

});

const noop = () => false;

function FormElement(props) {
  const {
    path,
    field,
    onChange
  } = props;
  const {
    data,
    errors,
    fields,
    properties,
    getFieldRenderer
  } = hooks.useContext(FormContext);
  const {
    Element
  } = hooks.useContext(FormRenderContext);
  const Renderer = getFieldRenderer(field.type);
  const {
    id
  } = field;

  if (!Renderer) {
    throw new Error(`cannot render field <${field.type}>`);
  }

  const value = findData(data, path);
  const fieldErrors = findErrors(errors, path);
  hooks.useEffect(() => {
    fields.set(id, { ...field,
      path
    });
  }, [pathStringify(path)]);
  return jsxRuntime.jsx(Element, {
    field: field,
    children: jsxRuntime.jsx(Renderer, { ...props,
      disabled: properties.readOnly || false,
      errors: fieldErrors,
      onChange: properties.readOnly ? noop : onChange,
      value: value
    })
  });
}

/**
 * This file must not be changed or exchanged.
 *
 * @see http://bpmn.io/license for more information.
 */

function Logo() {
  return jsxRuntime.jsxs("svg", {
    xmlns: "http://www.w3.org/2000/svg",
    viewBox: "0 0 14.02 5.57",
    width: "53",
    height: "21",
    style: "vertical-align:middle",
    children: [jsxRuntime.jsx("path", {
      fill: "currentColor",
      d: "M1.88.92v.14c0 .41-.13.68-.4.8.33.14.46.44.46.86v.33c0 .61-.33.95-.95.95H0V0h.95c.65 0 .93.3.93.92zM.63.57v1.06h.24c.24 0 .38-.1.38-.43V.98c0-.28-.1-.4-.32-.4zm0 1.63v1.22h.36c.2 0 .32-.1.32-.39v-.35c0-.37-.12-.48-.4-.48H.63zM4.18.99v.52c0 .64-.31.98-.94.98h-.3V4h-.62V0h.92c.63 0 .94.35.94.99zM2.94.57v1.35h.3c.2 0 .3-.09.3-.37v-.6c0-.29-.1-.38-.3-.38h-.3zm2.89 2.27L6.25 0h.88v4h-.6V1.12L6.1 3.99h-.6l-.46-2.82v2.82h-.55V0h.87zM8.14 1.1V4h-.56V0h.79L9 2.4V0h.56v4h-.64zm2.49 2.29v.6h-.6v-.6zM12.12 1c0-.63.33-1 .95-1 .61 0 .95.37.95 1v2.04c0 .64-.34 1-.95 1-.62 0-.95-.37-.95-1zm.62 2.08c0 .28.13.39.33.39s.32-.1.32-.4V.98c0-.29-.12-.4-.32-.4s-.33.11-.33.4z"
    }), jsxRuntime.jsx("path", {
      fill: "currentColor",
      d: "M0 4.53h14.02v1.04H0zM11.08 0h.63v.62h-.63zm.63 4V1h-.63v2.98z"
    })]
  });
}

function Lightbox(props) {
  const {
    open
  } = props;

  if (!open) {
    return null;
  }

  return jsxRuntime.jsxs("div", {
    class: "fjs-powered-by-lightbox",
    style: "z-index: 100; position: fixed; top: 0; left: 0;right: 0; bottom: 0",
    children: [jsxRuntime.jsx("div", {
      class: "backdrop",
      style: "width: 100%; height: 100%; background: rgba(40 40 40 / 20%)",
      onClick: props.onBackdropClick
    }), jsxRuntime.jsxs("div", {
      class: "notice",
      style: "position: absolute; left: 50%; top: 40%; transform: translate(-50%); width: 260px; padding: 10px; background: white; box-shadow: 0  1px 4px rgba(0 0 0 / 30%); font-family: Helvetica, Arial, sans-serif; font-size: 14px; display: flex; line-height: 1.3",
      children: [jsxRuntime.jsx("a", {
        href: "https://bpmn.io",
        target: "_blank",
        rel: "noopener",
        style: "margin: 15px 20px 15px 10px; align-self: center; color: #404040",
        children: jsxRuntime.jsx(Logo, {})
      }), jsxRuntime.jsxs("span", {
        children: ["Web-based tooling for BPMN, DMN, and forms powered by ", jsxRuntime.jsx("a", {
          href: "https://bpmn.io",
          target: "_blank",
          rel: "noopener",
          children: "bpmn.io"
        }), "."]
      })]
    })]
  });
}

function Link(props) {
  return jsxRuntime.jsx("div", {
    class: "fjs-powered-by fjs-form-field",
    style: "text-align: right",
    children: jsxRuntime.jsx("a", {
      href: "https://bpmn.io",
      target: "_blank",
      rel: "noopener",
      class: "fjs-powered-by-link",
      title: "Powered by bpmn.io",
      style: "color: #404040",
      onClick: props.onClick,
      children: jsxRuntime.jsx(Logo, {})
    })
  });
}

function PoweredBy(props) {
  const [open, setOpen] = hooks.useState(false);

  function toggleOpen(open) {
    return event => {
      event.preventDefault();
      setOpen(open);
    };
  }

  return jsxRuntime.jsxs(preact.Fragment, {
    children: [compat.createPortal(jsxRuntime.jsx(Lightbox, {
      open: open,
      onBackdropClick: toggleOpen(false)
    }), document.body), jsxRuntime.jsx(Link, {
      onClick: toggleOpen(true)
    })]
  });
}

const noop$1 = () => {};

function Form(props) {
  const {
    schema
  } = hooks.useContext(FormContext);
  const {
    onSubmit = noop$1,
    onReset = noop$1,
    onChange = noop$1
  } = props;

  const handleSubmit = event => {
    event.preventDefault();
    onSubmit();
  };

  const handleReset = event => {
    event.preventDefault();
    onReset();
  };

  return jsxRuntime.jsxs("form", {
    class: "fjs-form",
    onSubmit: handleSubmit,
    onReset: handleReset,
    children: [jsxRuntime.jsx(FormElement, {
      field: schema,
      onChange: onChange,
      path: []
    }), jsxRuntime.jsx(PoweredBy, {})]
  });
}

function formFieldClasses(type, errors = []) {
  if (!type) {
    throw new Error('type required');
  }

  const classes = ['fjs-form-field', `fjs-form-field-${type}`];

  if (errors.length) {
    classes.push('fjs-has-errors');
  }

  return classes.join(' ');
}
function prefixId(id) {
  return `fjs-form-${id}`;
}
const NODE_TYPE_TEXT = 3,
      NODE_TYPE_ELEMENT = 1;
const DISALLOWED_NODES = ['embed', 'iframe', 'object', 'script', 'svg'];
const ALLOWED_ATTRIBUTES = ['align', 'alt', 'class', 'href', 'id', 'name', 'src', 'valign']; // See https://github.com/developit/snarkdown/issues/70

function safeMarkdown(markdown) {
  const html = snarkdown__default['default'](markdown);
  const doc = new DOMParser().parseFromString(`<!DOCTYPE html>\n<html><body><div>${html}`, 'text/html');
  doc.normalize();
  sanitize(doc.body);
  return new XMLSerializer().serializeToString(doc.body.firstChild);
}

function sanitize(node) {
  if (node.nodeType === NODE_TYPE_TEXT) {
    return;
  }

  if (node.nodeType !== NODE_TYPE_ELEMENT || DISALLOWED_NODES.includes(node.tagName.toLowerCase())) {
    return node.remove();
  }

  for (let i = node.attributes.length; i--;) {
    const name = node.attributes[i].name;

    if (!ALLOWED_ATTRIBUTES.includes(name.toLowerCase())) {
      node.attributes.removeNamedItem(name);
    }

    if (name === 'href') {
      const href = node.attributes.getNamedItem('href');

      if (href.value.includes('javascript:')) {
        node.attributes.removeNamedItem('href');
      }
    }
  }

  for (let i = node.childNodes.length; i--;) {
    sanitize(node.childNodes[i]);
  }
}

const type = 'button';
function ButtonRenderer(props) {
  const {
    disabled,
    field
  } = props;
  const {
    action = 'submit'
  } = field;
  return jsxRuntime.jsx("div", {
    class: formFieldClasses(type),
    children: jsxRuntime.jsx("button", {
      class: "fjs-button",
      type: action,
      disabled: disabled,
      children: field.label
    })
  });
}

ButtonRenderer.create = function (options = {}) {
  const id = generateIdForType(type);
  return {
    action: 'submit',
    id,
    key: id,
    label: this.label,
    type,
    ...options
  };
};

ButtonRenderer.type = type;
ButtonRenderer.label = 'Button';

function Description(props) {
  const {
    description
  } = props;

  if (!description) {
    return null;
  }

  return jsxRuntime.jsx("div", {
    class: "fjs-form-field-description",
    children: description
  });
}

function Errors(props) {
  const {
    errors
  } = props;

  if (!errors.length) {
    return null;
  }

  return jsxRuntime.jsx("div", {
    class: "fjs-form-field-error",
    children: jsxRuntime.jsx("ul", {
      children: errors.map(error => {
        return jsxRuntime.jsx("li", {
          children: error
        });
      })
    })
  });
}

function Label(props) {
  const {
    id,
    label,
    required = false
  } = props;

  if (!label) {
    return null;
  }

  return jsxRuntime.jsxs("label", {
    for: id,
    class: "fjs-form-field-label",
    children: [props.children, label, required && jsxRuntime.jsx("span", {
      class: "fjs-asterix",
      children: "*"
    })]
  });
}

const type$1 = 'checkbox';
function CheckboxRenderer(props) {
  const {
    disabled,
    errors = [],
    field,
    path,
    value = false
  } = props;
  const {
    description,
    id,
    label
  } = field;

  const onChange = ({
    target
  }) => {
    props.onChange({
      path,
      value: target.checked
    });
  };

  return jsxRuntime.jsxs("div", {
    class: formFieldClasses(type$1, errors),
    children: [jsxRuntime.jsx(Label, {
      id: prefixId(id),
      label: label,
      required: false,
      children: jsxRuntime.jsx("input", {
        checked: value,
        class: "fjs-input",
        disabled: disabled,
        id: prefixId(id),
        type: "checkbox",
        onChange: onChange
      })
    }), jsxRuntime.jsx(Description, {
      description: description
    }), jsxRuntime.jsx(Errors, {
      errors: errors
    })]
  });
}

CheckboxRenderer.create = function (options = {}) {
  const id = generateIdForType(type$1);
  return {
    id,
    key: id,
    label: this.label,
    type: type$1,
    ...options
  };
};

CheckboxRenderer.type = type$1;
CheckboxRenderer.label = 'Checkbox';

function DefaultRenderer(props) {
  const {
    Children,
    Empty
  } = hooks.useContext(FormRenderContext);
  let {
    field,
    path
  } = props;
  const {
    id
  } = field;
  const {
    components = []
  } = field;
  return jsxRuntime.jsxs(Children, {
    class: "fjs-vertical-layout",
    field: field,
    children: [components.map(field => {
      if (field.key) {
        path = [...path.slice(0, -1), field.key];
      } else {
        path = path.slice(0, -1);
      }

      return preact.createElement(FormElement, { ...props,
        key: id,
        path: path,
        field: field
      });
    }), components.length ? null : jsxRuntime.jsx(Empty, {})]
  });
}

DefaultRenderer.create = function (options = {}) {
  const id = generateIdForType(this.type);
  return {
    components: [],
    id,
    label: this.label,
    type: this.type,
    ...options
  };
};

DefaultRenderer.type = 'default';
DefaultRenderer.label = 'Default';

const type$2 = 'number';
function NumberRenderer(props) {
  const {
    disabled,
    errors = [],
    field,
    path,
    value
  } = props;
  const {
    description,
    id,
    label,
    validate = {}
  } = field;
  const {
    required
  } = validate;

  const onChange = ({
    target
  }) => {
    props.onChange({
      path,
      value: target.value ? parseInt(target.value, 10) : undefined
    });
  };

  return jsxRuntime.jsxs("div", {
    class: formFieldClasses(type$2, errors),
    children: [jsxRuntime.jsx(Label, {
      id: prefixId(id),
      label: label,
      required: required
    }), jsxRuntime.jsx("input", {
      class: "fjs-input",
      disabled: disabled,
      id: prefixId(id),
      onInput: onChange,
      type: "number",
      value: value
    }), jsxRuntime.jsx(Description, {
      description: description
    }), jsxRuntime.jsx(Errors, {
      errors: errors
    })]
  });
}

NumberRenderer.create = function (options = {}) {
  const id = generateIdForType(type$2);
  return {
    id,
    key: id,
    label: this.label,
    type: type$2,
    ...options
  };
};

NumberRenderer.type = type$2;
NumberRenderer.label = 'Number';

const type$3 = 'radio';
function RadioRenderer(props) {
  const {
    disabled,
    errors = [],
    field,
    path,
    value
  } = props;
  const {
    description,
    id,
    label,
    validate = {},
    values
  } = field;
  const {
    required
  } = validate;

  const onChange = v => {
    props.onChange({
      path,
      value: v === value ? undefined : v
    });
  };

  return jsxRuntime.jsxs("div", {
    class: formFieldClasses(type$3, errors),
    children: [jsxRuntime.jsx(Label, {
      label: label,
      required: required
    }), values.map((v, index) => {
      return jsxRuntime.jsx(Label, {
        id: prefixId(`${id}-${index}`),
        label: v.label,
        required: false,
        children: jsxRuntime.jsx("input", {
          checked: v.value === value,
          class: "fjs-input",
          disabled: disabled,
          id: prefixId(`${id}-${index}`),
          type: "radio",
          onClick: () => onChange(v.value)
        })
      });
    }), jsxRuntime.jsx(Description, {
      description: description
    }), jsxRuntime.jsx(Errors, {
      errors: errors
    })]
  });
}

RadioRenderer.create = function (options = {}) {
  const id = generateIdForType(type$3);
  return {
    id,
    key: id,
    label: this.label,
    type: type$3,
    values: [{
      label: 'Value',
      value: 'value'
    }],
    ...options
  };
};

RadioRenderer.type = type$3;
RadioRenderer.label = 'Radio';

const type$4 = 'select';
function SelectRenderer(props) {
  const {
    disabled,
    errors = [],
    field,
    path,
    value
  } = props;
  const {
    description,
    id,
    label,
    validate = {},
    values
  } = field;
  const {
    required
  } = validate;

  const onChange = ({
    target
  }) => {
    props.onChange({
      path,
      value: target.value === '' ? undefined : target.value
    });
  };

  return jsxRuntime.jsxs("div", {
    class: formFieldClasses(type$4, errors),
    children: [jsxRuntime.jsx(Label, {
      label: label,
      required: required
    }), jsxRuntime.jsxs("select", {
      class: "fjs-select",
      disabled: disabled,
      id: prefixId(id),
      onChange: onChange,
      value: value,
      children: [jsxRuntime.jsx("option", {
        value: ""
      }), values.map((v, index) => {
        return jsxRuntime.jsx("option", {
          value: v.value,
          children: v.label
        });
      })]
    }), jsxRuntime.jsx(Description, {
      description: description
    }), jsxRuntime.jsx(Errors, {
      errors: errors
    })]
  });
}

SelectRenderer.create = function (options = {}) {
  const id = generateIdForType(type$4);
  return {
    id,
    key: id,
    label: this.label,
    type: type$4,
    values: [{
      label: 'Value',
      value: 'value'
    }],
    ...options
  };
};

SelectRenderer.type = type$4;
SelectRenderer.label = 'Select';

const type$5 = 'text';
function TextRenderer(props) {
  const {
    field
  } = props;
  const {
    text
  } = field;
  return jsxRuntime.jsx("div", {
    class: formFieldClasses(type$5),
    children: jsxRuntime.jsx(Markup__default['default'], {
      markup: safeMarkdown(text)
    })
  });
}

TextRenderer.create = function (options = {}) {
  const id = generateIdForType(type$5);
  return {
    id,
    text: '# Text',
    type: type$5,
    ...options
  };
};

TextRenderer.type = type$5;

const type$6 = 'textfield';
function TextfieldRenderer(props) {
  const {
    disabled,
    errors = [],
    field,
    path,
    value = ''
  } = props;
  const {
    description,
    id,
    label,
    validate = {}
  } = field;
  const {
    required
  } = validate;

  const onChange = ({
    target
  }) => {
    props.onChange({
      path,
      value: target.value.length ? target.value : undefined
    });
  };

  return jsxRuntime.jsxs("div", {
    class: formFieldClasses(type$6, errors),
    children: [jsxRuntime.jsx(Label, {
      id: prefixId(id),
      label: label,
      required: required
    }), jsxRuntime.jsx("input", {
      class: "fjs-input",
      disabled: disabled,
      id: prefixId(id),
      onInput: onChange,
      type: "text",
      value: value
    }), jsxRuntime.jsx(Description, {
      description: description
    }), jsxRuntime.jsx(Errors, {
      errors: errors
    })]
  });
}

TextfieldRenderer.create = function (options = {}) {
  const id = generateIdForType(type$6);
  return {
    id,
    key: id,
    label: this.label,
    type: type$6,
    ...options
  };
};

TextfieldRenderer.type = type$6;
TextfieldRenderer.label = 'Text Field';

const fields = [ButtonRenderer, CheckboxRenderer, // ColumnsRenderer,
DefaultRenderer, NumberRenderer, RadioRenderer, SelectRenderer, TextRenderer, TextfieldRenderer]; // export { createRenderer } from './CustomRenderer';

function FormRenderer(options) {
  const {
    additionalRenderers = [],
    container,
    form
  } = options;
  const renderers = [...additionalRenderers, ...fields];

  const App = () => {
    const [state, setState] = hooks.useState(form.getState());
    const formContext = {
      get fields() {
        return form.fields;
      },

      get data() {
        return state.data;
      },

      get errors() {
        return state.errors;
      },

      get properties() {
        return state.properties;
      },

      get schema() {
        return state.schema;
      },

      getFieldRenderer(type) {
        return findFieldRenderer(renderers, type);
      }

    };
    form.on('changed', newState => {
      setState(newState);
    });
    const onChange = hooks.useCallback(update => form.update(update), [form]);
    const {
      properties
    } = state;
    const {
      readOnly
    } = properties;
    const onSubmit = hooks.useCallback(() => {
      if (!readOnly) {
        form.submit();
      }
    }, [form, readOnly]);
    const onReset = hooks.useCallback(() => form.reset(), [form]);
    return jsxRuntime.jsx("div", {
      class: "fjs-container",
      children: jsxRuntime.jsx(FormContext.Provider, {
        value: formContext,
        children: jsxRuntime.jsx(Form, {
          onChange: onChange,
          onSubmit: onSubmit,
          onReset: onReset
        })
      })
    });
  };

  preact.render(jsxRuntime.jsx(App, {}), container);
}

const schemaVersion = 1;
/**
 * @typedef { { container: Element; schema: any; data: any; properties?: any } } FormViewerOptions
 */

/**
 * Create a form.
 *
 * @param {FormViewerOptions} options
 *
 * @return {FormCore}
 */

function createForm(options) {
  const {
    container,
    schema,
    data,
    properties = {}
  } = options;
  const form = new FormCore({
    schema,
    data,
    properties
  });
  new FormRenderer({
    container,
    form
  });
  return form;
}

exports.ButtonRenderer = ButtonRenderer;
exports.CheckboxRenderer = CheckboxRenderer;
exports.DefaultRenderer = DefaultRenderer;
exports.Form = Form;
exports.FormContext = FormContext;
exports.FormCore = FormCore;
exports.FormRenderContext = FormRenderContext;
exports.FormRenderer = FormRenderer;
exports.NumberRenderer = NumberRenderer;
exports.RadioRenderer = RadioRenderer;
exports.SelectRenderer = SelectRenderer;
exports.TextRenderer = TextRenderer;
exports.TextfieldRenderer = TextfieldRenderer;
exports.clone = clone;
exports.createForm = createForm;
exports.fields = fields;
exports.findData = findData;
exports.findErrors = findErrors;
exports.findFieldRenderer = findFieldRenderer;
exports.generateIdForType = generateIdForType;
exports.importSchema = importSchema;
exports.isRequired = isRequired;
exports.pathParse = pathParse;
exports.pathStringify = pathStringify;
exports.pathsEqual = pathsEqual;
exports.schemaVersion = schemaVersion;
//# sourceMappingURL=index.cjs.map
